# cython: boundscheck=False
# cython: cdivision=True
# cython: wraparound=False
from __future__ import division

import numpy as np

cimport numpy as np
from libc.math cimport cos, sin, atan2
cimport cython
ctypedef np.float64_t FLOAT_t


cdef struct st_rgb:
    double r
    double g
    double b

ctypedef st_rgb rgb

cdef struct st_lch:
    double l
    double c
    double h

ctypedef st_lch lch


def rgb_to_lch(r, g, b):
    """Convert RGB colors to LCH

    Parameters
    ----------
    r: float
        0 to 1, Red
    g: float
        0 to 1, Blue
    b: float
        0 to 1, Blue

    Returns
    -------
    L, C, H tuple (Lightness, Chroma, Hue)
    H is in radians
    """
    cdef lch color
    color = _rgb_to_lch(r, g, b)
    return color.l, color.c, color.h


def lch_to_rgb(l, c, h):
    """Convert LCH colors to RGB

    Parameters
    ----------
    L: float
        Lightness
    C: float
        Chroma
    H: float
        Hue, radians

    Returns
    -------
    r, g, b tuple (Red, Green and Blue)
    scaled float 0..1
    """
    cdef rgb color
    color = _lch_to_rgb(l, c, h)
    return color.r, color.g, color.b


cpdef np.ndarray[FLOAT_t, ndim=3] saturate_rgb(np.ndarray[FLOAT_t, ndim=3] arr, double satmult):
    """Convert array of RGB -> LCH, adjust saturation, back to RGB
    """
    cdef double r, g, b
    cdef size_t i, j
    cdef lch c_lch
    cdef rgb c_rgb
    if arr.shape[0] != 3:
        raise ValueError("The 0th dimension must contain 3 bands")

    cdef np.ndarray[FLOAT_t, ndim=3] out = np.empty_like(arr)

    for i in range(arr.shape[1]):
        for j in range(arr.shape[2]):
            r = arr[0, i, j]
            g = arr[1, i, j]
            b = arr[2, i, j]

            c_lch = _rgb_to_lch(r, g, b)
            c_lch.c *= satmult
            c_rgb = _lch_to_rgb(c_lch.l, c_lch.c, c_lch.h)

            out[0, i, j] = c_rgb.r
            out[1, i, j] = c_rgb.g
            out[2, i, j] = c_rgb.b

    return out


cpdef np.ndarray[FLOAT_t, ndim=3] arr_rgb_to_lch(np.ndarray[FLOAT_t, ndim=3] arr):
    cdef double r, g, b
    cdef size_t i, j
    cdef lch color
    if arr.shape[0] != 3:
        raise ValueError("The 0th dimension must contain 3 bands")

    cdef np.ndarray[FLOAT_t, ndim=3] out = np.empty_like(arr)

    for i in range(arr.shape[1]):
        for j in range(arr.shape[2]):
            r = arr[0, i, j]
            g = arr[1, i, j]
            b = arr[2, i, j]
            color = _rgb_to_lch(r, g, b)
            out[0, i, j] = color.l
            out[1, i, j] = color.c
            out[2, i, j] = color.h

    return out


cpdef np.ndarray[FLOAT_t, ndim=3] arr_lch_to_rgb(np.ndarray[FLOAT_t, ndim=3] arr):
    cdef double l, c, h
    cdef size_t i, j
    cdef rgb color
    if arr.shape[0] != 3:
        raise ValueError("The 0th dimension must contain 3 bands")

    cdef np.ndarray[FLOAT_t, ndim=3] out = np.empty_like(arr)

    for i in range(arr.shape[1]):
        for j in range(arr.shape[2]):
            l = arr[0, i, j]
            c = arr[1, i, j]
            h = arr[2, i, j]
            color = _lch_to_rgb(l, c, h)
            out[0, i, j] = color.r
            out[1, i, j] = color.g
            out[2, i, j] = color.b

    return out

# Constants
cdef double bintercept = 4.0 / 29  # 0.137931
cdef double delta = 6.0 / 29  # 0.206896
cdef double t0 = delta ** 3  # 0.008856
cdef double alpha = (delta ** -2) / 3  # 7.787037


cdef lch _rgb_to_lch(double r, double g, double b):
    cdef lch color

    # convert RGB to linear scale
    # if simplified:
    # Use gamma = 2.2, "simplified sRGB"
    rl = r ** 2.2
    gl = g ** 2.2
    bl = b ** 2.2

    # matrix mult for srgb->xyz,
    # includes adjustment for d65 reference white
    x = ((rl * 0.4124) + (gl * 0.3576) + (bl * 0.1805)) / 0.95047
    y = ((rl * 0.2126) + (gl * 0.7152) + (bl * 0.0722))
    z = ((rl * 0.0193) + (gl * 0.1192) + (bl * 0.9505)) / 1.08883

    # convert XYZ to LAB colorspace
    if x > t0:
        fx = x ** 0.3333333333333333
    else:
        fx = (alpha * x) + bintercept

    if y > t0:
        fy = y ** 0.3333333333333333
    else:
        fy = (alpha * y) + bintercept

    if z > t0:
        fz = z ** 0.3333333333333333
    else:
        fz = (alpha * z) + bintercept

    color.l = (116 * fy) - 16
    a = 500 * (fx - fy)
    b = 200 * (fy - fz)

    color.c = ((a * a) + (b * b)) ** 0.5
    color.h = atan2(b, a)

    return color


cdef rgb _lch_to_rgb(double L, double C, double H):
    cdef double a, b
    cdef rgb color

    a = C * cos(H)
    b = C * sin(H)

    tx = ((L + 16) / 116.0) + (a / 500.0)
    if tx > delta:
        x = tx ** 3
    else:
        x = 3 * delta * delta * (tx - bintercept)

    ty = (L + 16) / 116.0
    if ty > delta:
        y = ty ** 3
    else:
        y = 3 * delta * delta * (ty - bintercept)

    tz = ((L + 16) / 116.0) - (b / 200.0)
    if tz > delta:
        z = tz ** 3
    else:
        z = 3 * delta * delta * (tz - bintercept)

    # uses reference white d65
    x = x * 0.95047
    z = z * 1.08883

    # XYZ to sRGB
    # expanded matrix multiplication
    rlin = (x * 3.2406) + (y * -1.5372) + (z * -0.4986)
    glin = (x * -0.9689) + (y * 1.8758) + (z * 0.0415)
    blin = (x * 0.0557) + (y * -0.2040) + (z * 1.0570)

    # constrain to 0..1 to deal with any float drift
    if rlin > 1.0:
        rlin = 1.0
    elif rlin < 0.0:
        rlin = 0.0
    if glin > 1.0:
        glin = 1.0
    elif glin < 0.0:
        glin = 0.0
    if blin > 1.0:
        blin = 1.0
    elif blin < 0.0:
        blin = 0.0

    # includes gamma exponentiation
    # Use simplified sRGB with gamma = 2.2
    color.r = rlin ** (1 / 2.2)
    color.g = glin ** (1 / 2.2)
    color.b = blin ** (1 / 2.2)

    return color



import click

import rasterio
from rasterio.rio.options import creation_options
from rio_color.workers import atmos_worker, color_worker
from rio_color.operations import parse_operations, simple_atmo_opstring
import riomucho


jobs_opt = click.option(
    '--jobs', '-j', type=int, default=1,
    help="Number of jobs to run simultaneously, Use -1 for all cores, default: 1")


def check_jobs(jobs):
    if jobs == 0:
        raise click.UsageError("Jobs must be >= 1 or == -1")
    elif jobs < 0:
        import multiprocessing
        jobs = multiprocessing.cpu_count()
    return jobs


@click.command('color')
@jobs_opt
@click.option('--out-dtype', '-d', type=click.Choice(['uint8', 'uint16']),
              help="Integer data type for output data, default: same as input")
@click.argument('src_path', type=click.Path(exists=True))
@click.argument('dst_path', type=click.Path(exists=False))
@click.argument('operations', nargs=-1, required=True)
@click.pass_context
@creation_options
def color(ctx, jobs, out_dtype, src_path, dst_path, operations,
          creation_options):
    """Color correction

Operations will be applied to the src image in the specified order.
Each operation should be a single quoted argument.

Available OPERATIONS include:

\b
    "gamma BANDS VALUE"
        Applies a gamma curve, brightening or darkening midtones.
        VALUE > 1 brightens the image.

\b
    "sigmoidal BANDS CONTRAST BIAS"
        Adjusts the contrast and brightness of midtones.
        BIAS > 0.5 darkens the image.

\b
    "saturation PROPORTION"
        Controls the saturation in LCH color space.
        PROPORTION = 0 results in a grayscale image
        PROPORTION = 1 results in an identical image
        PROPORTION = 2 is likely way too saturated

BANDS are specified as a single arg

\b
    `123` or `RGB` or `rgb` are all equivalent

Example:

\b
    rio color -d uint8 -j 4 input.tif output.tif \\
        gamma 3 0.95 sigmoidal 1,2,3 35 0.13
    """
    with rasterio.open(src_path) as src:
        opts = src.profile.copy()
        windows = [(window, ij) for ij, window in src.block_windows()]

    opts.update(**creation_options)
    opts['transform'] = opts['affine']

    out_dtype = out_dtype if out_dtype else opts['dtype']
    opts['dtype'] = out_dtype

    args = {
        'ops_string': ' '.join(operations),
        'out_dtype': out_dtype
    }
    # Just run this for validation this time
    # parsing will be run again within the worker
    # where its returned value will be used
    try:
        ops = parse_operations(args['ops_string'])
    except ValueError as e:
        raise click.UsageError(str(e))

    jobs = check_jobs(jobs)

    if jobs > 1:
        with riomucho.RioMucho(
            [src_path],
            dst_path,
            color_worker,
            windows=windows,
            options=opts,
            global_args=args,
            mode="manual_read"
        ) as mucho:
            mucho.run(jobs)
    else:
        with rasterio.open(dst_path, 'w', **opts) as dest:
            with rasterio.open(src_path) as src:
                rasters = [src]
                for window, ij in windows:
                    arr = color_worker(rasters, window, ij, args)
                    dest.write(arr, window=window)


@click.command('atmos')
@click.option('--atmo', '-a', type=click.FLOAT, default=0.03,
              help="How much to dampen cool colors, thus cutting through "
                   "haze. 0..1 (0 is none), default: 0.03.")
@click.option('--contrast', '-c', type=click.FLOAT, default=10,
              help="Contrast factor to apply to the scene. -infinity..infinity"
                   "(0 is none), default: 10.")
@click.option('--bias', '-b', type=click.FLOAT, default=0.15,
              help="Skew (brighten/darken) the output. Lower values make it "
                   "brighter. 0..1 (0.5 is none), default: 0.15")
@click.option('--out-dtype', '-d', type=click.Choice(['uint8', 'uint16']),
              help="Integer data type for output data, default: same as input")
@click.option('--as-color', is_flag=True, default=False,
              help="Prints the equivalent rio color command to stdout."
                   "Does NOT run either command, SRC_PATH will not be created")
@click.argument('src_path', required=True)
@click.argument('dst_path', type=click.Path(exists=False))
@jobs_opt
@creation_options
@click.pass_context
def atmos(ctx, atmo, contrast, bias, jobs, out_dtype,
          src_path, dst_path, creation_options, as_color):
    """Atmospheric correction
    """
    if as_color:
        click.echo("rio color {} {} {}".format(
            src_path, dst_path, simple_atmo_opstring(atmo, contrast, bias)))
        exit(0)

    with rasterio.open(src_path) as src:
        opts = src.profile.copy()
        windows = [(window, ij) for ij, window in src.block_windows()]

    opts.update(**creation_options)
    opts['transform'] = opts['affine']

    out_dtype = out_dtype if out_dtype else opts['dtype']
    opts['dtype'] = out_dtype

    args = {
        'atmo': atmo,
        'contrast': contrast,
        'bias': bias,
        'out_dtype': out_dtype
    }

    jobs = check_jobs(jobs)

    if jobs > 1:
        with riomucho.RioMucho(
            [src_path],
            dst_path,
            atmos_worker,
            windows=windows,
            options=opts,
            global_args=args,
            mode="manual_read"
        ) as mucho:
            mucho.run(jobs)
    else:
        with rasterio.open(dst_path, 'w', **opts) as dest:
            with rasterio.open(src_path) as src:
                rasters = [src]
                for window, ij in windows:
                    arr = atmos_worker(rasters, window, ij, args)
                    dest.write(arr, window=window)



from .operations import parse_operations, simple_atmo
from .utils import to_math_type, scale_dtype

# Rio workers

def atmos_worker(srcs, window, ij, args):
    src = srcs[0]
    rgb = src.read(window=window)
    rgb = to_math_type(rgb)

    atmos = simple_atmo(
        rgb,
        args['atmo'],
        args['contrast'],
        args['bias'])

    # should be scaled 0 to 1, scale to outtype
    return scale_dtype(atmos, args['out_dtype'])


def color_worker(srcs, window, ij, args):
    src = srcs[0]
    arr = src.read(window=window)
    arr = to_math_type(arr)

    for func in parse_operations(args['ops_string']):
        arr = func(arr)

    # scaled 0 to 1, now scale to outtype
    return scale_dtype(arr, args['out_dtype'])



import os
import sys
from setuptools import setup, find_packages
from setuptools.extension import Extension

# Use Cython if available.
try:
    from Cython.Build import cythonize
except ImportError:
    cythonize = None

include_dirs = []
try:
    import numpy
    include_dirs.append(numpy.get_include())
except ImportError:
    print("Numpy and its headers are required to run setup(). Exiting.")
    sys.exit(1)


# Parse the version from the fiona module.
with open('rio_color/__init__.py') as f:
    for line in f:
        if line.find("__version__") >= 0:
            version = line.split("=")[1].strip()
            version = version.strip('"')
            version = version.strip("'")
            break

long_description = """Color adjustment plugin for rasterio.

See https://github.com/mapbox/rio-color for docs."""


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

if cythonize and 'clean' not in sys.argv:
    ext_modules = cythonize([
        Extension(
            "rio_color.colorspace", ["rio_color/colorspace.pyx"])])
else:
    ext_modules = [
        Extension(
            "rio_color.colorspace", ["rio_color/colorspace.c"])]

setup(name='rio-color',
      version=version,
      description=u"Color correction plugin for rasterio",
      long_description=long_description,
      classifiers=[],
      keywords='',
      author=u"Charlie Loyd",
      author_email='charlie@mapbox.com',
      url='https://github.com/mapbox/rio-color',
      license='BSD',
      packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
      include_package_data=True,
      zip_safe=False,
      install_requires=["click", "rasterio", "rio-mucho"],
      ext_modules=ext_modules,
      include_dirs=include_dirs,
      extras_require={
          'test': ['pytest', 'pytest-cov', 'codecov', 'raster-tester'],
      },
      entry_points="""
      [rasterio.rio_plugins]
      color=rio_color.scripts.cli:color
      atmos=rio_color.scripts.cli:atmos
      """
      )
